from sys import (
    stdout,
)


def function_tools_after_migrate_receiver(sender, **kwargs):
    """
    Действия выполняемые после прогона миграций.
    """
    from function_tools.management.commands.register_entities import (
        EntityRegistrar,
    )

    registrar = EntityRegistrar(logger=stdout)
    registrar.run()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2022-07-24 00:50
from __future__ import unicode_literals

import django.contrib.postgres.fields
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('function_tools', '0004_implementationstrategy_order_number'),
    ]

    operations = [
        migrations.CreateModel(
            name='Entity',
            fields=[
                ('uuid', models.UUIDField(primary_key=True, serialize=False, unique=True, verbose_name='Уникальный идентификатор')),
                ('class_name', models.CharField(max_length=256, verbose_name='Название класса')),
                ('import_path', models.CharField(max_length=512, verbose_name='Путь до модуля для импорта')),
                ('verbose_name', models.CharField(default='Имя сущности не определено', max_length=512, verbose_name='Человеко читаемое название')),
                ('tags', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(blank=True, max_length=128), default=list, size=None)),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Дата и время создания')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Дата и время обновления')),
            ],
            options={
                'verbose_name': 'Сущность',
                'verbose_name_plural': 'Сущности',
                'db_table': 'function_tools_entity',
            },
        ),
        migrations.CreateModel(
            name='EntityType',
            fields=[
                ('title', models.TextField(verbose_name='расшифровка значения')),
                ('key', models.CharField(db_index=True, max_length=512, primary_key=True, serialize=False, verbose_name='ключ')),
                ('order_number', models.PositiveIntegerField(default=100000, verbose_name='Порядковый номер')),
            ],
            options={
                'verbose_name': 'Тип сущности',
                'verbose_name_plural': 'Типы сущностей',
                'db_table': 'function_tools_entity_type',
            },
        ),
        migrations.DeleteModel(
            name='RegisteredFunction',
        ),
        migrations.AddField(
            model_name='entity',
            name='type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='function_tools.EntityType', verbose_name='Тип сущности'),
        ),
    ]



import warnings

from function_tools.strategies import (
    FunctionImplementationStrategy,
    SyncBaseRunnerBaseFunctionImplementationStrategy,
    SyncBaseRunnerLazySavingPredefinedQueueFunctionImplementationStrategy,
    SyncLazySavingRunnerLazyDelegateSavingPredefinedQueueFunctionImplementationStrategy,
)


warnings.warn(
    message=(
        'Модуль function_tools.management.strategies был перемещен function_tools.strategies. Просьба внести '
        'исправления в импорты.'
    ),
    category=DeprecationWarning,
    stacklevel=2,
)



from inspect import (
    isclass,
)
from pathlib import (
    Path,
)
from typing import (
    Optional,
    Type,
    Union,
)

from django.contrib.postgres.fields import (
    ArrayField,
)
from django.db.models import (
    CharField,
    DateTimeField,
    ForeignKey,
    Model,
    UUIDField,
)

from function_tools.functions import (
    BaseFunction,
)
from function_tools.managers import (
    RunnerManager,
)
from function_tools.runners import (
    BaseRunner,
)
from function_tools.strategies import (
    FunctionImplementationStrategy,
)
from m3_db_utils.models import (
    ModelEnumValue,
    TitledModelEnum,
)


class ImplementationStrategy(TitledModelEnum):
    """
    Перечисление стратегий реализации функций.
    """

    class Meta:
        db_table = 'function_tools_implementation_strategy'
        verbose_name = 'Стратегия создания функции'
        extensible = True


class EntityType(TitledModelEnum):
    """
    Тип сущности function_tools.
    """

    MANAGER = ModelEnumValue(
        title='Менеджер',
        module='managers',
        base_class=RunnerManager,
    )

    RUNNER = ModelEnumValue(
        title='Ранер',
        module='runners',
        base_class=BaseRunner,
    )

    FUNCTION = ModelEnumValue(
        title='Функция',
        module='functions',
        base_class=BaseFunction,
    )

    STRATEGY = ModelEnumValue(
        title='Стратегия создания Функции',
        module='strategies',
        base_class=FunctionImplementationStrategy,
    )

    class Meta:
        db_table = 'function_tools_entity_type'
        verbose_name = 'Тип сущности'
        verbose_name_plural = 'Типы сущностей'

    @classmethod
    def get_type(
        cls,
        entity: Union[object, Type[object]],
    ) -> Optional[ModelEnumValue]:
        """
        Определение типа сущности. Может передаваться как класс, так и экземпляр класса.

        Args:
            entity: Класс или экземпляр класса
        """
        entity_type = None

        if isclass(entity):
            check_method = issubclass
        else:
            check_method = isinstance

        for temp_entity_type in cls.get_model_enum_values():
            if check_method(entity, temp_entity_type.base_class):
                entity_type = temp_entity_type
                break

        return entity_type

    @classmethod
    def get_type_by_path(cls, module_path: Union[str, Path]):
        """
        Определение типа сущности по пути модуля.

        Args:
            module_path: Путь модуля
        """
        module_path = Path(module_path)

        module_name = module_path.name.split('.')[0]

        entity_type = None

        for temp_entity_type in cls.get_model_enum_values():
            if temp_entity_type.module == module_name:
                entity_type = temp_entity_type
                break

        return entity_type


class Entity(Model):
    """
    Сущность function_tools хранящаяся в базе данных.
    """

    uuid = UUIDField(
        verbose_name='Уникальный идентификатор',
        primary_key=True,
        unique=True,
    )

    class_name = CharField(
        verbose_name='Название класса',
        max_length=256,
    )

    import_path = CharField(
        verbose_name='Путь до модуля для импорта',
        max_length=512,
    )

    verbose_name = CharField(
        verbose_name='Человеко читаемое название',
        max_length=512,
        default='Имя сущности не определено',
    )

    tags = ArrayField(
        base_field=CharField(max_length=128, blank=True),
        default=list,
    )

    type = ForeignKey(
        to=EntityType,
        verbose_name='Тип сущности',
    )

    created_at = DateTimeField(
        verbose_name='Дата и время создания',
        auto_now_add=True,
    )
    updated_at = DateTimeField(
        verbose_name='Дата и время обновления',
        auto_now=True,
    )

    class Meta:
        db_table = 'function_tools_entity'
        verbose_name = 'Сущность'
        verbose_name_plural = 'Сущности'



import sys
from typing import (
    Dict,
    List,
    Optional,
    Type,
    Union,
)

from django.core.management import (
    color_style,
)
from django.core.management.base import (
    BaseCommand,
)
from django.db import (
    transaction,
)

from function_tools.models import (
    Entity,
    EntityType,
)
from function_tools.storages import (
    EntityStorage,
)
from function_tools.management.strategies import SyncBaseRunnerBaseFunctionImplementationStrategy


class EntityRegistrar:
    """
    Регистратор классов сущностей системы.
    """

    compared_fields = {
        'class_name': '__name__',
        'verbose_name': 'verbose_name',
        'tags': 'tags',
    }

    def __init__(self, logger):
        self._logger = logger

        self._entities: Optional[Dict[str, Dict[str, Union[Type[object], str]]]] = None

        self._registered_entities: Optional[List[Entity]] = None
        self._registered_entities_map: Optional[Dict[str, Entity]] = None

        self._style = color_style()

        self._prepare_registered_entities()
        self._prepare_entities()

    def _prepare_registered_entities(self):
        """
        Подготовка зарегистрированных сущностей в базе данных.
        """
        self._registered_entities = list(Entity.objects.all())

        self._registered_entities_map = {
            str(registered_entity.uuid): registered_entity
            for registered_entity in self._registered_entities
        }

    def _prepare_entities(self):
        """
        Поиск реализованных классов сущностей.
        """
        entity_storage = EntityStorage()
        entity_storage.prepare()

        self._entities = entity_storage.flat_entities

    def _process_creating_entities(self):
        """
        Обработка создаваемых классов сущностей.
        """
        self._logger.write('start processing creating entities..\n')

        for_creating = []

        creating_entity_uuids = filter(
            lambda uuid: uuid not in self._registered_entities_map.keys(),
            self._entities.keys()
        )

        for creating_entity_uuid in creating_entity_uuids:
            entity = self._entities[creating_entity_uuid]

            for_creating.append(
                Entity(
                    uuid=creating_entity_uuid,
                    verbose_name=entity['class'].verbose_name,
                    tags=entity['class'].tags,
                    type_id=EntityType.get_type(entity['class']).key,
                    class_name=entity['class'].__name__,
                    import_path=entity['import_path'],
                )
            )

        if for_creating:
            Entity.objects.bulk_create(objs=for_creating)

        self._logger.write('finished processing creating entities.\n')

    def _compare_entities(
        self,
        registered_entity: Entity,
        runtime_entity: Type[object],
        runtime_entity_import_path: str,
    ) -> List[str]:
        """
        Производит сравнение значений полей двух экземпляров сущности из базы и из кода. Возвращает список полей для
        обновления.
        """
        updated_fields = []

        for model_field_name, class_field_name in self.compared_fields.items():
            if getattr(registered_entity, model_field_name) != getattr(runtime_entity, class_field_name):
                updated_fields.append(model_field_name)

        if registered_entity.import_path != runtime_entity_import_path:
            updated_fields.append('import_path')

        return updated_fields

    def _process_updating_entities(self):
        """
        Обработка обновляемых классов сущностей.
        """
        self._logger.write('start processing updating entities..\n')

        updating_entities_uuids = filter(
            lambda uuid: uuid in self._registered_entities_map,
            self._entities.keys()
        )

        for updating_entity_uuid in updating_entities_uuids:
            updated_fields = self._compare_entities(
                registered_entity=self._registered_entities_map[updating_entity_uuid],
                runtime_entity=self._entities[updating_entity_uuid]['class'],
                runtime_entity_import_path=self._entities[updating_entity_uuid]['import_path'],
            )

            if updated_fields:
                for updated_field in updated_fields:
                    if updated_field == 'import_path':
                        self._registered_entities_map[updating_entity_uuid].import_path = (
                            self._entities[updating_entity_uuid]['import_path']
                        )
                    else:
                        setattr(
                            self._registered_entities_map[updating_entity_uuid],
                            updated_field,
                            getattr(self._entities[updating_entity_uuid]['class'], self.compared_fields[updated_field])
                        )

                    # Иду на преступный шаг с генерацией некоторого количества запросов в БД, все только для того, чтобы
                    # изменить дату и время обновления записи
                    self._registered_entities_map[updating_entity_uuid].save()

        self._logger.write('processing updating entities finished.\n')

    def _process_deleting_entities(self):
        """
        Обработка удаленных, перенесенных или находящихся в отключенных плагинах сущностях системы.
        """
        self._logger.write('start processing deleting entities..\n')

        deleted_entities_uuids = set(
            self._registered_entities_map.keys()
        ).difference(self._entities.keys())

        if deleted_entities_uuids:
            for deleted_entity_uuid in deleted_entities_uuids:
                sys.stdout.write(
                    self._style.NOTICE(
                        f'Found registered Entity in the database, but not found implementation. Entity with UUID "'
                        f'{deleted_entity_uuid}". Please, check it!\n'
                    )
                )

        self._logger.write('end processing deleting entities..\n')

    def _register(self):
        """
        Регистрация реализованных сущностей системы.
        """
        self._process_creating_entities()
        self._process_updating_entities()
        self._process_deleting_entities()

    def run(self):
        self._logger.write('start function_tools entities registration..\n')

        self._register()

        self._logger.write('function_tools entities registration finished.\n')


class Command(BaseCommand):
    """
    Команда регистрации сущностей системы.
    """

    help = 'Команда предназначена для регистрации сущностей системы, реализованных при помощи function-tools.'

    @transaction.atomic
    def handle(self, *args, **kwargs):
        registrar = EntityRegistrar(logger=self.stdout)
        registrar.run()



import sys
from collections import (
    defaultdict,
)
from importlib import (
    import_module,
)
from inspect import (
    isclass,
)
from pathlib import (
    Path,
)
from types import (
    ModuleType,
)
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
)

from django.apps import (
    AppConfig,
)
from django.conf import (
    settings,
)

from function_tools.functions import (
    BaseFunction,
)
from function_tools.managers import (
    RunnerManager,
)
from function_tools.models import (
    EntityType,
)
from function_tools.runners import (
    BaseRunner,
)
from m3_db_utils.models import (
    ModelEnumValue,
)


if TYPE_CHECKING:
    from django.apps import (
        AppConfig,
    )


class EntityStorage:
    """
    Хранилище классов сущностей реализованных в системе.

    Собираются только те сущности, типы которых указаны в модели-перечислении
    function_tools.models.function_tools.models.EntityType.
    """

    def __init__(self):
        # Словарь для хранения найденных классов реализованных сущностей
        self._entities = {
            model_enum_value.key: defaultdict()
            for model_enum_value in EntityType.get_model_enum_values()
        }

        self._checked_application_paths: List[Path] = []

        self._entity_module_patterns: List[str] = [
            f'**/**/{model_enum_value.module}.py'
            for model_enum_value in EntityType.get_model_enum_values()
        ]

        self._entities_modules: List[Tuple[Path, str, ModuleType]] = []

        self._sys_path = set(sys.path)

    @property
    def entities(self) -> Dict[str, Dict[str, Dict[str, Union[str, Type[object]]]]]:
        """
        Все классы сущностей реализованных в системе, сгруппированные по типу сущностям.
        """
        return self._entities

    @property
    def flat_entities(self) -> Dict[str, Dict[str, Union[Type[object], str]]]:
        """
        Возвращает плоский словарь с классами сущностей с UUID в качестве ключа.
        """
        flat_entities = {}

        for entities in self._entities.values():
            flat_entities.update(entities)

        return flat_entities

    def _get_module_import_path(self, module_path: str) -> str:
        """
        Предназначен для получения пути пакета для импорта класса сущности.

        Args:
            module_path: абсолютный путь до модуля
        """
        package_path = max(
            filter(
                lambda path: path in module_path,
                self._sys_path
            )
        )

        relative_module_path = module_path.split(f'{package_path}/')[1]
        import_path = '.'.join(relative_module_path.split('.')[0].split('/'))

        return import_path

    def _get_application_path(self, application_name: Union[str, 'AppConfig']) -> Optional[Path]:
        """
        Получение пути приложения.

        Args:
            application_name: Наименование приложения
        """
        # TODO EDUSCHL-17934 В app_name могут прилетать AppConfig. Необходимо произвести доработку и обработать
        #  такие случаи.
        try:
            app_module = import_module(application_name)
        except ModuleNotFoundError:
            return

        application_path = Path(app_module.__path__[0])

        # Если поиск уже осуществлялся по родительской директории,
        # то проверку нужно пропустить
        is_already_checked = False

        for checked_application_path in self._checked_application_paths:
            try:
                is_already_checked = application_path.is_relative_to(Path(checked_application_path))
            # В CI падает с ошибкой 'PosixPath' object has no attribute 'is_relative_to', хотя локально не
            # воспроизводится
            except AttributeError:
                continue

            if is_already_checked:
                break

        return application_path if not is_already_checked else None

    def _find_application_entities_modules(self, application_name: Union[str, 'AppConfig']):
        """
        Поиск модулей зарегистрированных типов сущностей в приложении.

        Args:
            application_name: имя приложения
        """
        application_path = self._get_application_path(application_name=application_name)

        if application_path:
            for entity_module_pattern in self._entity_module_patterns:
                entity_module_paths = application_path.glob(entity_module_pattern)

                for entity_module_path in entity_module_paths:
                    entity_module_path = str(entity_module_path)

                    import_path = self._get_module_import_path(
                        module_path=entity_module_path,
                    )

                    try:
                        import_module(import_path)

                        entity_module = sys.modules[import_path]
                    except (KeyError, RuntimeError):
                        continue

                    self._entities_modules.append((Path(entity_module_path), import_path, entity_module))

            self._checked_application_paths.append(application_path)

    def _find_entities_modules(self):
        """
        Поиск модулей зарегистрированных типов сущностей.
        """
        for application_name in settings.INSTALLED_APPS:
            self._find_application_entities_modules(application_name=application_name)

    def _prepare_entities(self):
        """
        Поиск фабрик во всех подключенных приложениях.
        """
        processed_paths = list()

        for entity_module_path, entity_module_import_path, entity_module in self._entities_modules:
            if entity_module_path in processed_paths:
                continue
            else:
                processed_paths.append(entity_module_path)

            entity_type: ModelEnumValue = EntityType.get_type_by_path(module_path=entity_module_path)

            for item_name, item in entity_module.__dict__.items():
                if (
                    isclass(item)
                    and issubclass(item, entity_type.base_class)
                    and hasattr(item, 'uuid')
                    and item.uuid
                ):
                    if item.uuid in self.flat_entities.keys() and item != self.flat_entities[item.uuid]['class']:
                        raise RuntimeError(f'Found duplicated UUID {item.uuid}! Please check and change it!')

                    self._entities[entity_type.key][item.uuid] = {
                        'class': item,
                        'import_path': entity_module_import_path,
                    }

    def prepare_runner_function_map(
        self,
        runner_class: Type[BaseRunner],
        runner_tags: Optional[Set[str]] = None,
        function_tags: Optional[Set[str]] = None,
    ) -> Tuple[Type[BaseRunner], List]:
        """
        Собирает и возвращает список функций запускаемых в рамках работы менеджера.
        """
        entities = []

        for runnable_class in runner_class._prepare_runnable_classes():
            if issubclass(runnable_class, BaseRunner):
                if runner_tags:
                    if not runner_tags.difference(runnable_class.tags):
                        entities.append(
                            self.prepare_runner_function_map(
                                runner_class=runnable_class,
                            )
                        )
                else:
                    entities.append(
                        self.prepare_runner_function_map(
                            runner_class=runnable_class,
                        )
                    )

                continue
            elif issubclass(runnable_class, BaseFunction):
                if function_tags:
                    if not function_tags.difference(runnable_class.tags):
                        entities.append(runnable_class)
                else:
                    entities.append(runnable_class)

        return runner_class, entities

    def prepare_manager_runnable_map(
        self,
        manager_tags: Optional[Set[str]] = None,
        runner_tags: Optional[Set[str]] = None,
        function_tags: Optional[Set[str]] = None,
    ) -> Dict[Type[RunnerManager], Tuple[Type[BaseRunner], List]]:
        """
        Строит карту соответствия менеджер -> ранер -> функция.

        Args:
            manager_tags: теги менеджеров;
            runner_tags: теги ранеров;
            function_tags: теги функций.
        """
        manager_runner_function_map = dict()

        manager_classes: List[Type[RunnerManager]] = []
        for entity in self.entities[EntityType.MANAGER.key].values():
            if manager_tags:
                if not manager_tags.difference(entity['class'].tags):
                    manager_classes.append(entity['class'])
            else:
                manager_classes.append(entity['class'])

        for manager_class in manager_classes:
            runner_class = manager_class.runner_class

            manager_runner_function_map[manager_class] = self.prepare_runner_function_map(
                runner_class=runner_class,
                runner_tags=runner_tags,
                function_tags=function_tags,
            )

        return manager_runner_function_map

    def _find_functions(self, entities, function_uuids):
        """
        Поиск функций в списке сущностей, который может содержать вложенный список.
        """
        for entity in entities:
            if isinstance(entity, tuple):
                _, temp_entities = entity

                self._find_functions(
                    entities=temp_entities,
                    function_uuids=function_uuids,
                )
            elif issubclass(entity, BaseFunction):
                function_uuids.append(entity.uuid)

    def prepare_manager_function_uuid_map(
        self,
        manager_tags: Optional[Set[str]] = None,
        runner_tags: Optional[Set[str]] = None,
        function_tags: Optional[Set[str]] = None,
    ) -> Dict[str, List[str]]:
        """
        Возвращает карту соответствия менеджера и функций.

        В качестве ключей и значений списка используются UUID-ы сущностей.
        """
        manager_function_uuid_map = {}

        manager_runnable_map = self.prepare_manager_runnable_map(
            manager_tags=manager_tags,
            runner_tags=runner_tags,
            function_tags=function_tags,
        )

        for manager_class, (_, entities) in manager_runnable_map.items():
            function_uuids = []

            self._find_functions(
                entities=entities,
                function_uuids=function_uuids,
            )

            manager_function_uuid_map[manager_class.uuid] = function_uuids

        return manager_function_uuid_map

    def prepare(self):
        """
        Подготовка хранилища. Производится его наполнение реализованными классами сущностей.
        """
        self._find_entities_modules()
        self._prepare_entities()
